/*

Flexible Arrays untersttzen Subskription in `log n' und knnen zustzlich
in konstanter Zeit vorne erweitert werden (amortisierte Laufzeit unter der
Voraussetzung, dass die Datenstruktur ephemer verwendet wird).

*/

module Flexible-Array =
struct
  data Array <a> = Null
                 | Zero    (Array <(a, a)>)
                 | One  (a, Array <(a, a)>)

  // Konstruktion flexibler Arrays.

  function nil <a : Type> : Array <a> = Null

  function cons <a> (x : a, list : Array <a>) : Array <a> =
    case list of
      Null        => One (x, Null)
    | Zero xs     => One (x, xs)
    | One (y, xs) => Zero (cons <(a, a)> ((x, y), xs))
    end

  function flexible-array <a> (n : Nat, map : Nat -> a) : Array <a> =
    if n == 0 then
      Null
    else if n % 2 == 0 then 
      Zero (       flexible-array <(a, a)> (n / 2, fun i => (map (2 * i),     map (2 * i + 1))))
    else 
      One  (map 0, flexible-array <(a, a)> (n / 2, fun i => (map (2 * i + 1), map (2 * i + 2))))

  // Subskription.

  function index <a> (list : Array <a>, i : Nat) : a =
    case list of
      Null        => throw Index
    | Zero xs     => if i % 2 == 0 then (index <(a, a)> (xs, i / 2)).1
                                   else (index <(a, a)> (xs, i / 2)).2
    | One (x, xs) => if i == 0 then x else index <a> (Zero xs, i - 1)
    end

  // Konversion (Listen in flexibler Arrays und umgekehrt).

  local
    open List

    function unpairs <a> (list : List <(a, a)>) : List <a> =
      case list of
        Nil               => Nil
      | Cons ((x, y), xs) => Cons (x, Cons (y, unpairs <a> xs)	)
      end
  in
    function from-list <a> (list : List <a>) : Array <a> =
      fold <a, Array <a>> (nil <a>, cons <a>) list

    function to-list <a> (list : Array <a>) : List <a> =
      case list of
        Null        => Nil
      | Zero xs     => unpairs <a> (to-list <(a, a)> xs)
      | One (x, xs) => Cons (x, unpairs <a> (to-list <(a, a)> xs))
      end
  end
end // Flexible-Array
